/* $Id: cilk-internal.h,v 1.1.1.1.2.2 1997/09/06 06:22:37 fengmd Exp $ */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

/* common definitions used internally by the rts */

/******************************************************
 *  Closures and associated operations
 ******************************************************/
/*
 * If a frame is at the top of its processor stack, these additional
 * informations are recorded.
 */
enum ClosureStatus { CLOSURE_RUNNING = 42, CLOSURE_SUSPENDED,
                     CLOSURE_RETURNING, CLOSURE_READY };

/* The order is important here. */
enum AbortStatus { ABORT_ALL = 30 , ALMOST_NO_ABORT, NO_ABORT};

/* 
 * the list of children is not distributed among
 * the children themselves, in order to avoid extra protocols
 * and locking.
 */
/*
 * This structure contains all the data needed in order to execute an
 * inlet.  Some inlets are "complete", other are not.  "Incomplete"
 * inlets are still waiting for data.
 *
 * Each closure has a list of complete inlets, and a list
 * of incomplete inlets.  The list of incomplete inlets
 * is isomorphic to the list of outstanding children,
 * and is also used for that purpose.
 * Complete inlets are ready to go, but are waiting for
 * atomic access to the frame.
 */
struct InletClosure {
     struct Closure *this;                /* child that structure is for */
     struct InletClosure *next;
  
     int index;
     void (*inlet) (void *, void *, void *);
     int argsize;
     void *inlet_args;
};

typedef struct Closure {
     Cilk_mutex mutex;          /* mutual exclusion lock */
     WHEN_CILK_DEBUG(int mutex_owner;)

     int join_counter; /* number of outstanding children */
     enum ClosureStatus status;
     
     /*
      * return value (and its size), valid only 
      * for CLOSURE_RETURNING closures 
      */
     void *return_value; 
     int return_size;

     struct InletClosure *complete_inlets;  /* runnable inlets */
     struct InletClosure *incomplete_inlets;  
     /* also, list of outstanding children */

     struct Closure *parent; /* the parent closure */

     StackFrame *frame;       /* rest of the closure */
     enum AbortStatus abort_status; 
     struct ClosureCache *cache; /* see below */

     /* 
      * stuff related to ready deque.  These fields
      * must be managed only by the queue manager in sched.c
      */
     struct Closure *next_ready; 
     struct Closure *prev_ready; 
     WHEN_CILK_DEBUG(int owner_ready_deque;)

     /* miscellanea */
     WHEN_CILK_DEBUG(int magic;)

     /* critical path and work */
     Cilk_time cp;      
     Cilk_time work;

     CILK_CACHE_LINE_PAD;
} Closure;

/* internal malloc */
extern void Cilk_internal_malloc_per_processor_init(void);
extern void Cilk_internal_malloc_global_init(void);
extern void Cilk_internal_malloc_print_statistics(void);

/* barriers */
extern void Cilk_barrier(void);
extern void Cilk_barrier_init(void);

/* scheduler */
extern void Cilk_scheduler_init(void);
extern Closure *Cilk_create_initial_thread(int argc, char *argv[]);
extern void Cilk_scheduler(Closure *t);

/* other variables */
DECLARE_SHARED_RO_VARIABLE(extern int, yield_slice, );
DECLARE_SHARED_RO_VARIABLE(extern int, StatLevel, );
DECLARE_SHARED_RO_VARIABLE(extern int, trace_mode, );
DECLARE_SHARED_RO_VARIABLE(extern int, dump_core, );
DECLARE_SHARED_VARIABLE(extern int, retval, );

/* external malloc */
extern void Cilk_init_arena(size_t arenasize);

/* statistics */
#if CILK_STATS

enum { 
     EVENT_THREAD,

     EVENT_STEAL_ATTEMPT,
     EVENT_STEAL,
     EVENT_STEAL_NO_DEKKER,
     EVENT_STEAL_EMPTY_DEQUE,
     EVENT_STEAL_RETURNING,
     EVENT_STEAL_ABORT,

     EVENT_ABORT_SLOW,
     EVENT_ABORT_STANDALONE,
     EVENT_ABORT_READY_SUSPEND,
     EVENT_ABORT_READY_RETURN,

     EVENT_SUSPEND,
     EVENT_CILK_SYNC,

     EVENT_EXCEPTION,
     EVENT_EXCEPTION_STEAL,
     EVENT_EXCEPTION_ABORT,
     EVENT_EXCEPTION_ABORT_RETURN,
     EVENT_EXCEPTION_ABORT_SUSPEND,
     EVENT_EXCEPTION_OTHER,

     EVENT_RETURN_SLOW,
     EVENT_PROVABLY_GOOD_STEAL,
     EVENT_RETURN_ENQUEUE,
     EVENT_POLL_INLETS,
     EVENT_CLOSURE_CREATE,

     EVENT_USER0,
     EVENT_USER1,
     EVENT_USER2,
     EVENT_USER3,
     EVENT_USER4,
     EVENT_USER5,
     EVENT_USER6,
     EVENT_USER7,
     EVENT_NTYPES  		/* this must be last */
};
     
extern void Cilk_event(int type);
extern void Cilk_event_gathering_init(void);
extern void Cilk_print_rts_statistics(void);

#else

#define Cilk_event(a)
#define Cilk_event_gathering_init()
#define Cilk_print_rts_statistics()

#endif

/* timing */
#if CILK_TIMING
extern void Cilk_enter_state(int state);
extern void Cilk_exit_state(int state);
extern void Cilk_print_time_statistics(void);
extern void Cilk_print_detailed_time_statistics(void);
extern void Cilk_summarize_time_statistics(void);
extern double Cilk_compute_work(void);
extern void Cilk_time_gathering_init(void);

enum {
     STATE_TOTAL,
     STATE_WORKING,
     STATE_WAITING_FOR_LOCK,
     STATE_STEALING,
     STATE_RETURNING,
     STATE_ABORT_RECURSIVE,
     STATE_USER0,                /* for the user */
     STATE_USER1,
     STATE_USER2,
     STATE_USER3,
     STATE_USER4,
     STATE_USER5,
     STATE_USER6,
     STATE_USER7,
     TIME_NSTATES		/* this must be last */
};

#else

#define Cilk_enter_state(state)
#define Cilk_exit_state(state)
#define Cilk_print_time_statistics()
#define Cilk_print_detailed_time_statistics()
#define Cilk_summarize_time_statistics()
#define Cilk_compute_work() (0.0)
#define Cilk_time_gathering_init()

#endif				/* timing */

/*
 * Functions defined in the architecture-specific file and used
 * only by the runtime system.
 */
extern void Cilk_map_shared_segment(void);
extern void Cilk_map_private_segment(void);
extern void *Cilk_map_shared(size_t len);
extern int Cilk_arch_specific_options(int i, int argc, char *argv[]);
extern void Cilk_arch_specific_init(void);
extern void Cilk_arch_specific_per_process_init(void);
extern void Cilk_create_children(void (*child)(int));
extern void Cilk_lower_priority(void);
extern void Cilk_raise_priority(void);

/*
 * functions used by the scheduler and by other protocols
 * operating on the Closures tree
 */
extern Closure *Cilk_Closure_create(void);


/******************************************************
 *  Options processing
 ******************************************************/

typedef struct
{
     int nproc;
     int stackdepth;
     size_t heap;
     int statlevel;
     int yieldslice;
     char *infofile_name;  /* this should really be arch-specific */
     int dump_core;
} Options;

#define DEFAULT_OPTIONS   (Options)           \
{ 					      \
     1, 				      \
     CILK_DEFAULT_STACK_DEPTH, 		      \
     CILK_DEFAULT_HEAPSIZE, 		      \
     0, 				      \
     1, 				      \
     "-", 				      \
     0					      \
}

/* command-line parser */
extern int Cilk_parse_command_line(Options *options, int *argc, char *argv[]);

