/*
 * $Id: internal-malloc.c,v 1.1.1.1.2.6 1997/09/06 06:22:42 fengmd Exp $
 */

/*
 * $Log: internal-malloc.c,v $
 * Revision 1.1.1.1.2.6  1997/09/06 06:22:42  fengmd
 * Merge change from the main trunk.
 *
 * Revision 1.1.1.1.2.5  1997/09/02 12:09:47  fengmd
 * Minor change to remove fprintf in internal-malloc().
 *
 * Revision 1.1.1.1.2.4  1997/08/29 07:40:18  fengmd
 * Merge malloc/free checker, race report hashing, garbage collection.
 *
 * Revision 1.1.1.1.2.3  1997/06/16 16:00:47  fengmd
 * Remerge from the base for Version 5.0.1.
 *
 * Revision 1.2  1997/03/01 17:42:53  athena
 * Massive cleanup and merging --- since I was still using the
 * old repository :-(
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */
#include <cilk.h>
#include <cilk-lib.h>
#include <cilk-internal.h>
#include <cilk-cilk2c.h>
#include <stdio.h>
#include <string.h>

DECLARE_SHARED_RO_VARIABLE(extern FILE *, infofile, );

/*
 * This is a memory allocator that is optimized for small pieces
 * of data.  Here is how it works.
 *
 * There is a global pool of memory (managed by Cilk_malloc()).  When
 * Cilk_internal_malloc(size) is executed, there are two cases. If
 * size is large, the normal Cilk_malloc() allocates memory, as
 * needed.  Otherwise, internal_malloc() manages the memory.  It
 * allocates an entire chunk of memory (using Cilk_malloc()) and
 * handles it privately on a per-processor basis.
 */

#define CILK_INTERNAL_MALLOC_CHUNK (32 * 1024)

DECLARE_PRIVATE_VARIABLE(static char *, pool_begin, );
DECLARE_PRIVATE_VARIABLE(static char *, pool_end, );

/* WARNING: not properly initialized */
DECLARE_PRIVATE_VARIABLE(void *, im_free_lists, [CILK_INTERNAL_MALLOC_FREELISTS]);
/*
 * pointer to an array containing the algebraic sum
 * of the memory used by each processor.  The sum
 * of the array must be 0 at the end of the program.
 */
DECLARE_SHARED_RO_VARIABLE(struct Cilk_im_stats *, im_info, );

static void allocate_private_pool(void)
{
     USE_PRIVATE(pool_begin) = Cilk_malloc_fixed(CILK_INTERNAL_MALLOC_CHUNK);
     CILK_CHECK(USE_PRIVATE(pool_begin),
		("internal malloc() out of memory\n"));
     USE_PRIVATE(pool_end) = USE_PRIVATE(pool_begin) +
	  CILK_INTERNAL_MALLOC_CHUNK;
     USE_SHARED_RO(im_info)[Self].allocated += 
	  CILK_INTERNAL_MALLOC_CHUNK;
}

void *Cilk_internal_malloc(size_t size)
{
     int bucket;
     void *mem;
     WHEN_CILK_ND(size_t origsize = size;)

     WHEN_CILK_DEBUG(USE_SHARED_RO(im_info)[Self].used += size);

     if (size > CILK_INTERNAL_MALLOC_THRESHOLD)
	  return Cilk_malloc(size);

     size = Cilk_internal_malloc_canonicalize(size);

     bucket = Cilk_internal_malloc_size_to_bucket(size);

     /* look at the free list */
     if ((mem = USE_PRIVATE(im_free_lists)[bucket])) {
	  CILK_ASSERT(mem);
	  USE_PRIVATE(im_free_lists)[bucket] = *((void **) mem);
     } else {
	  /* else, allocate from the private pool */
	  if (USE_PRIVATE(pool_begin) + size > USE_PRIVATE(pool_end)) {
	       USE_SHARED_RO(im_info)[Self].wasted += 
		    (USE_PRIVATE(pool_begin) + size - USE_PRIVATE(pool_end));
	       allocate_private_pool();
	  }
	  mem = USE_PRIVATE(pool_begin);
	  USE_PRIVATE(pool_begin) += size;
     }

#ifdef CILK_ND
     {
       DisjointSetMemberT *hr, *hw;
       int i;
       origsize = (((origsize + sizeof(DisjointSetMemberT) - 1) &
		    (~(sizeof(DisjointSetMemberT) -1))) /
		   sizeof(DisjointSetMemberT));
       hr = HEAP_READER(mem);
       hw = HEAP_WRITER(mem);
       for (i=0;i<origsize;i++) {
	 update_ref_count(hr+i, *(hr+i), USE_SHARED_RO(current_proc_id),
			  USE_SHARED_RO(current_proc_id));
	 update_ref_count(hw+i, PROC_ID(*(hw+i)), 
			  USE_SHARED_RO(current_proc_id),
			  USE_SHARED_RO(current_proc_id));
       }
     }
#endif

     return mem;
}

/*
 * free simply returns to the free list 
 */
void Cilk_internal_free(void *p, size_t size)
{
     int bucket;

     WHEN_CILK_DEBUG(USE_SHARED_RO(im_info)[Self].used -= size);

     if (size > CILK_INTERNAL_MALLOC_THRESHOLD)
	  return Cilk_free(p);

#ifdef CILK_ND
     {
       int more_size = Cilk_internal_malloc_canonicalize(size);
       DisjointSetMemberT *hr, *hw;
       int i;

       /* internal_free can't cause a determinacy race, so just write 
          FREE_ID into the shadow space */
       CILK_ASSERT((more_size & (sizeof(DisjointSetMemberT) -1)) == 0);
       more_size /= sizeof(DisjointSetMemberT);
       hr = HEAP_READER(p);
       hw = HEAP_WRITER(p);
       for (i=0;i<more_size;i++) {
	 update_ref_count(hr+i, *(hr+i), FREE_ID, FREE_ID);
	 update_ref_count(hw+i, PROC_ID(*(hw+i)), FREE_ID, FREE_ID);
       }
     }
#endif

     bucket = Cilk_internal_malloc_size_to_bucket(size);

     *((void **) p) = USE_PRIVATE(im_free_lists)[bucket];
     USE_PRIVATE(im_free_lists)[bucket] = p;
}

void Cilk_internal_malloc_per_processor_init(void)
{
     int i;

     for (i = 0; i < CILK_INTERNAL_MALLOC_FREELISTS; ++i)
	  USE_PRIVATE(im_free_lists)[i] = (void *) NULL;

     allocate_private_pool();
}

/*
 * support for alloca
 */
#if CILK_ALLOCA
DECLARE_PRIVATE_VARIABLE(extern struct ClosureCache *const, cache ,);

void *Cilk_alloca(size_t size) 
{
     StackFrame *f = (StackFrame *)(USE_PRIVATE(cache)->tail)[-1];
     struct alloca_header *p;

     size += sizeof(struct alloca_header);
     p = Cilk_internal_malloc(size);
     p->size = size;
     p->next = f->alloca_h;
     f->alloca_h = p;
     return (void *)(p+1);
}
#endif /* CILK_ALLOCA */

void Cilk_internal_malloc_global_init(void)
{
     int i;

     CILK_ASSERT(Self == -1);
     INIT_SHARED_RO(im_info,
		    Cilk_malloc_fixed(USE_SHARED_RO(active_size) *
				      sizeof(struct Cilk_im_stats)));
     CILK_CHECK(USE_SHARED_RO(im_info), ("Can't allocate Cilk_im_info\n"));

     for (i = 0; i < USE_SHARED_RO(active_size); ++i) {
	  USE_SHARED_RO(im_info)[i].used = 0;
	  USE_SHARED_RO(im_info)[i].wasted = 0;
	  USE_SHARED_RO(im_info)[i].allocated = 0;
     }
}

void Cilk_internal_malloc_final_check(void)
{
     int i, total;

     total = 0;
     for (i = 0; i < USE_SHARED_RO(active_size); ++i)
	  total += USE_SHARED_RO(im_info)[i].used;

     CILK_COMPLAIN(total == 0,
		   ("Cilk runtime system: Cilk detected a memory leak.\n"
		    "Either this is a cilk bug (unlikely), or your program\n"
		    "exited without giving Cilk a chance to free its own\n"
		    "memory.  Make sure you use Cilk_exit() and not exit().\n"));
}

#define FORMAT "%14d%11d\n"
void Cilk_internal_malloc_print_statistics(void)
{
     int i;
     int allocated = 0;
     int wasted = 0;

     fprintf(USE_SHARED_RO(infofile),
	     "\nSCHEDULER MEMORY STATISTICS:\n"
	     "\n"
	     " PN     Allocated     Wasted\n"
	     "----------------------------\n");

     for (i = 0; i < USE_SHARED_RO(active_size); ++i) {
	  fprintf(USE_SHARED_RO(infofile), "%3d" FORMAT, i,
		  USE_SHARED_RO(im_info)[i].allocated,
		  USE_SHARED_RO(im_info)[i].wasted);

	  allocated += USE_SHARED_RO(im_info)[i].allocated;
	  wasted += USE_SHARED_RO(im_info)[i].wasted;
     }

     fprintf(USE_SHARED_RO(infofile),
	     "----------------------------\n");

     fprintf(USE_SHARED_RO(infofile), "   " FORMAT, allocated, wasted);
}
