/*
 * $Id: timing.c,v 1.1.1.1.2.2 1997/06/16 16:00:55 fengmd Exp $
 */

/*
 * $Log: timing.c,v $
 * Revision 1.1.1.1.2.2  1997/06/16 16:00:55  fengmd
 * Remerge from the base for Version 5.0.1.
 *
 * Revision 1.2  1997/03/01 17:42:59  athena
 * Massive cleanup and merging --- since I was still using the
 * old repository :-(
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

/***********************************************************
 * Time-gathering stuff
 ***********************************************************/

#include <cilk.h>
#include <cilk-lib.h>
#include <cilk-internal.h>

#if CILK_TIMING
#include <stdio.h>
DECLARE_SHARED_RO_VARIABLE(extern FILE *, infofile, );

/*
 * The model for time gathering is the following:
 * there is a certain set of states in which the system can
 * be. Throughout the code, ENTER_STATE(something) signals
 * that a given state is entered. EXIT_STATE(state) reverts back to
 * the previous state, and checks that the right state is being
 * exited. The time is accumulated on a per-state
 * basis. Time is cumulative, i.e. it is measured from ENTER_STATE
 * to EXIT_STATE, regardless of other states entered in between.
 * 
 * No more than STATE_DEPTH states can be nested.
 * The stack is really used only for checking purposes, but it's
 * probably a good idea to keep it.
 *
 */
#define STATE_DEPTH 256

/* also update name table below */

typedef struct StateInfo {
     volatile int state_stackp;
     volatile int state_stack[STATE_DEPTH];
     volatile Cilk_time state_stack_begintime[STATE_DEPTH];
     volatile Cilk_time statetime[TIME_NSTATES];
     volatile Cilk_time state_stack_begin_wall_time[STATE_DEPTH];
     volatile Cilk_time state_wall_time[TIME_NSTATES];
     volatile int stateentered[TIME_NSTATES];
     CILK_CACHE_LINE_PAD;
} StateInfoT;

DECLARE_SHARED_RO_VARIABLE(static struct StateInfo *, timer_state_info, );

/*
 * table of description of states. It is printed in
 * the same order.
 */
static const struct state_desc {
     char *name;
     int state;
} state_descriptions[] = {

     {
	  "Total execution time", STATE_TOTAL
     },
     {
	  "Working time", STATE_WORKING
     },
     {
	  "Waiting for locks", STATE_WAITING_FOR_LOCK
     },
     {
	  "Stealing", STATE_STEALING
     },
     {
	  "Returning", STATE_RETURNING
     },
     {
	  "Aborting", STATE_ABORT_RECURSIVE
     },
     {
	  "User state 0", STATE_USER0
     },
     {
	  "User state 1", STATE_USER1
     },
     {
	  "User state 2", STATE_USER2
     },
     {
	  "User state 3", STATE_USER3
     },
     {
	  "User state 4", STATE_USER4
     },
     {
	  "User state 5", STATE_USER5
     },
     {
	  "User state 6", STATE_USER6
     },
     {
	  "User state 7", STATE_USER7
     },
     {
	  (char *) 0, 0
     }
};

void Cilk_enter_state(int state)
{
     int stackp;
     Cilk_time now = Cilk_get_time();
     Cilk_time now_wall = Cilk_get_wall_time();
     StateInfoT *info;
     if (USE_SHARED_RO(timer_state_info) == NULL || Self == -1)
	  return;
     info = USE_SHARED_RO(timer_state_info) + Self;
     stackp = info->state_stackp;
     CILK_ASSERT(stackp < STATE_DEPTH);
     info->state_stack[stackp] = state;
     info->state_stack_begintime[stackp] = now;
     info->state_stack_begin_wall_time[stackp] = now_wall;
     info->state_stackp = stackp + 1;
     ++(info->stateentered[state]);
}

void Cilk_exit_state(int state)
{
     int stackp;
     Cilk_time now = Cilk_get_time();
     Cilk_time now_wall = Cilk_get_wall_time();
     StateInfoT *info;
     if (USE_SHARED_RO(timer_state_info) == NULL || Self == -1)
	  return;
     info = USE_SHARED_RO(timer_state_info) + Self;
     stackp = info->state_stackp - 1;
     info->state_stackp = stackp;
     CILK_ASSERT(stackp >= 0);

     CILK_ASSERT(info->state_stack[stackp] == state);
     info->statetime[state] += now - info->state_stack_begintime[stackp];
     info->state_wall_time[state] += 
	  now_wall - info->state_stack_begin_wall_time[stackp];
}

/* print time statistics */
void Cilk_print_time_statistics(void)
{
     int i;
     StateInfoT *info = USE_SHARED_RO(timer_state_info);
     double worktime = 0.0;
     double totaltime = 0.0;

     fprintf(USE_SHARED_RO(infofile),
	     "\nSUMMARY OF TIME SPENT BY EACH PROCESSOR:\n\n"
	     " PN     Working       Total\n"
	     "---------------------------\n");
     for (i = 0; i < USE_SHARED_RO(active_size); i++) {
	  fprintf(USE_SHARED_RO(infofile), "%3d %11f %11f\n", i,
	      Cilk_time_to_sec(info[i].statetime[STATE_WORKING]),
	       Cilk_time_to_sec(info[i].statetime[STATE_TOTAL]));
	  worktime +=
	      Cilk_time_to_sec(info[i].statetime[STATE_WORKING]);
	  totaltime +=
	      Cilk_time_to_sec(info[i].statetime[STATE_TOTAL]);
     }
     fprintf(USE_SHARED_RO(infofile),
	     "---------------------------\n"
	     "TOT %11f %11f\n", worktime, totaltime);
}

void Cilk_summarize_time_statistics(void)
{
     int i;
     struct state_desc *p;
     StateInfoT *info = USE_SHARED_RO(timer_state_info);

     fprintf(USE_SHARED_RO(infofile),
	     "\nSUMMARY OF TIME STATISTICS:\n\n");

     for (p = (struct state_desc *) state_descriptions; p->name; ++p) {
	  Cilk_time virtual = 0;
	  Cilk_time wall = 0;
	  int entered = 0;

	  for (i = 0; i < USE_SHARED_RO(active_size); i++) {
	       virtual += info[i].statetime[p->state];
	       wall += info[i].state_wall_time[p->state];
	       entered += info[i].stateentered[p->state];
	  }

	  if (entered > 0)
	       fprintf(USE_SHARED_RO(infofile),
		       "%s: %fs / wall %fs (entered %d times)\n",
		       p->name,
		       Cilk_time_to_sec(virtual),
		       Cilk_wall_time_to_sec(wall),
		       entered);
     }
}

void Cilk_print_detailed_time_statistics(void)
{
     int i;
     struct state_desc *p;
     StateInfoT *info = USE_SHARED_RO(timer_state_info);

     fprintf(USE_SHARED_RO(infofile),
	     "\nDETAILS OF TIME SPENT BY EACH PROCESSOR:\n");

     for (i = 0; i < USE_SHARED_RO(active_size); i++) {
	  fprintf(USE_SHARED_RO(infofile),
		  "\n--- Time statistics for processor %d ---\n", i);

	  for (p = (struct state_desc *) state_descriptions; p->name; ++p) {
	       if (info[i].stateentered[p->state] > 0)
		    fprintf(USE_SHARED_RO(infofile),
			    "%s: %fs / wall %fs (entered %d times)\n",
			    p->name,
			    Cilk_time_to_sec(info[i].statetime
						     [p->state]),
			    Cilk_wall_time_to_sec(info[i].state_wall_time
						  [p->state]),
			    info[i].stateentered[p->state]);
	  }
     }
}

double Cilk_compute_work(void)
{
     int i;
     double work = 0.0;
     StateInfoT *info = USE_SHARED_RO(timer_state_info);

     for (i = 0; i < USE_SHARED_RO(active_size); i++)
	  work += Cilk_time_to_sec(info[i].statetime[STATE_WORKING]);

     return work;
}

void Cilk_time_gathering_init(void)
{
     int i, j;
     StateInfoT *info;

     CILK_ASSERT(Self == -1);
     INIT_SHARED_RO(timer_state_info ,
		    (StateInfoT *) Cilk_malloc_fixed(sizeof(StateInfoT) *
					   USE_SHARED_RO(active_size)));
     info = USE_SHARED_RO(timer_state_info);

     for (i = 0; i < USE_SHARED_RO(active_size); i++) {
	  for (j = 0; j < TIME_NSTATES; ++j) {
	       info[i].statetime[j] = (Cilk_time) 0;
	       info[i].stateentered[j] = 0;
	  }
	  info[i].state_stackp = 0;
     }
}

#endif
